// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// Copyright(C) 2000 James Haley
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//--------------------------------------------------------------------------
//
// DESCRIPTION:
//      Created by the sound utility written by Dave Taylor.
//      Kept as a sample, DOOM2 sounds. Frozen.
//
//      haleyjd: erm... ::lights a fire to warm it up::
//
//-----------------------------------------------------------------------------

#ifndef __SOUNDS__
#define __SOUNDS__

//
// SoundFX struct.
//

struct sfxinfo_s;

typedef struct sfxinfo_s sfxinfo_t;
typedef struct musicinfo_s musicinfo_t;

struct sfxinfo_s
{

  // up to 6-character name
  char *name;

  // Sfx singularity (only one at a time)
  // killough 12/98: implement separate classes of singularity
  enum
  {
    sg_none,
    sg_itemup,
    sg_wpnup,
    sg_oof,
    sg_getpow
  }
  singularity;

  // Sfx priority
  int priority;

  // referenced sound if a link
  sfxinfo_t *link;

  // pitch if a link
  int pitch;

  // volume if a link
  int volume;

  // sound data
  void *data;

        // sf: skin sound number to use in place
  int skinsound;

  // this is checked every second to see if sound
  // can be thrown out (if 0, then decrement, if -1,
  // then throw out, if > 0, then it is in use)
  int usefulness;

  int length;   // lump length

  sfxinfo_t *next;      // next in hash chain
};

//
// MusicInfo struct.
//

struct musicinfo_s
{
  // up to 6-character name
  char *name;

  // lump number of music
  //  int lumpnum;

  // music data
  void *data;

  // music handle once registered
  int handle;

  // sf: for hashing
  musicinfo_t *next;
  
};

// the complete set of sound effects
extern sfxinfo_t    S_sfx[];

// the complete set of music
extern musicinfo_t  S_music[];

//
// Identifiers for all music in game.
//

typedef enum {
  mus_None,
  mus_e1m1,
  mus_e1m2,
  mus_e1m3,
  mus_e1m4,
  mus_e1m5,
  mus_e1m6,
  mus_e1m7,
  mus_e1m8,
  mus_e1m9,
  mus_e2m1,
  mus_e2m2,
  mus_e2m3,
  mus_e2m4,
  mus_e2m5,
  mus_e2m6,
  mus_e2m7,
  mus_e2m8,
  mus_e2m9,
  mus_e3m1,
  mus_e3m2,
  mus_e3m3,
  mus_e3m4,
  mus_e3m5,
  mus_e3m6,
  mus_e3m7,
  mus_e3m8,
  mus_e3m9,
  mus_inter,
  mus_intro,
  mus_bunny,
  mus_victor,
  mus_introa,
  mus_runnin,
  mus_stalks,
  mus_countd,
  mus_betwee,
  mus_doom,
  mus_the_da,
  mus_shawn,
  mus_ddtblu,
  mus_in_cit,
  mus_dead,
  mus_stlks2,
  mus_theda2,
  mus_doom2,
  mus_ddtbl2,
  mus_runni2,
  mus_dead2,
  mus_stlks3,
  mus_romero,
  mus_shawn2,
  mus_messag,
  mus_count2,
  mus_ddtbl3,
  mus_ampie,
  mus_theda3,
  mus_adrian,
  mus_messg2,
  mus_romer2,
  mus_tense,
  mus_shawn3,
  mus_openin,
  mus_evil,
  mus_ultima,
  mus_read_m,
  mus_dm2ttl,
  mus_dm2int,
  NUMMUSIC
} musicenum_t;

//
// Identifiers for all sfx in game.
//

typedef enum {
  sfx_None,
  sfx_pistol,
  sfx_shotgn,
  sfx_sgcock,
  sfx_dshtgn,
  sfx_dbopn,
  sfx_dbcls,
  sfx_dbload,
  sfx_plasma,
  sfx_bfg,
  sfx_sawup,
  sfx_sawidl,
  sfx_sawful,
  sfx_sawhit,
  sfx_rlaunc,
  sfx_rxplod,
  sfx_firsht,
  sfx_firxpl,
  sfx_pstart,
  sfx_pstop,
  sfx_doropn,
  sfx_dorcls,
  sfx_stnmov,
  sfx_swtchn,
  sfx_swtchx,
  sfx_plpain,
  sfx_dmpain,
  sfx_popain,
  sfx_vipain,
  sfx_mnpain,
  sfx_pepain,
  sfx_slop,
  sfx_itemup,
  sfx_wpnup,
  sfx_oof,
  sfx_telept,
  sfx_posit1,
  sfx_posit2,
  sfx_posit3,
  sfx_bgsit1,
  sfx_bgsit2,
  sfx_sgtsit,
  sfx_cacsit,
  sfx_brssit,
  sfx_cybsit,
  sfx_spisit,
  sfx_bspsit,
  sfx_kntsit,
  sfx_vilsit,
  sfx_mansit,
  sfx_pesit,
  sfx_sklatk,
  sfx_sgtatk,
  sfx_skepch,
  sfx_vilatk,
  sfx_claw,
  sfx_skeswg,
  sfx_pldeth,
  sfx_pdiehi,
  sfx_podth1,
  sfx_podth2,
  sfx_podth3,
  sfx_bgdth1,
  sfx_bgdth2,
  sfx_sgtdth,
  sfx_cacdth,
  sfx_skldth,
  sfx_brsdth,
  sfx_cybdth,
  sfx_spidth,
  sfx_bspdth,
  sfx_vildth,
  sfx_kntdth,
  sfx_pedth,
  sfx_skedth,
  sfx_posact,
  sfx_bgact,
  sfx_dmact,
  sfx_bspact,
  sfx_bspwlk,
  sfx_vilact,
  sfx_noway,
  sfx_barexp,
  sfx_punch,
  sfx_hoof,
  sfx_metal,
  sfx_chgun,
  sfx_tink,
  sfx_bdopn,
  sfx_bdcls,
  sfx_itmbk,
  sfx_flame,
  sfx_flamst,
  sfx_getpow,
  sfx_bospit,
  sfx_boscub,
  sfx_bossit,
  sfx_bospn,
  sfx_bosdth,
  sfx_manatk,
  sfx_mandth,
  sfx_sssit,
  sfx_ssdth,
  sfx_keenpn,
  sfx_keendt,
  sfx_skeact,
  sfx_skesit,
  sfx_skeatk,
  sfx_radio,

#ifdef DOGS
  // killough 11/98: dog sounds
  sfx_dgsit,
  sfx_dgatk,
  sfx_dgact,
  sfx_dgdth,
  sfx_dgpain,
#endif

  // Start Eternity TC New SFX
  // haleyjd 1/22/99: Minotaurus sounds
  sfx_minsit,
  sfx_minat1,
  sfx_minat2,
  sfx_minat3,
  sfx_minpai,
  sfx_mindth,
  sfx_minact,
  sfx_stfpow,
  sfx_phohit, // 1/23/99: minotaur fx sounds
  // end minotaurus sounds
  // start ambient sounds
  sfx_brook1, // water
  sfx_winde,   // joel - prevent eterntiy conflict
  sfx_wind3,
  sfx_water1, // waterfall
  sfx_amb3,   // drip 1  
  sfx_amb7,   // drip 2
  sfx_amb9,   // laugh hoo
  sfx_amb10,  // laugh ha
  sfx_bird,   // bird caw
  sfx_bugs1,  // crickets
  sfx_bugs2,
  sfx_bugs3,  // cicadas
  sfx_ktydid, // katydid
  sfx_frogs,  // frogs
  sfx_owl,    // hoot owl
  // end ambient sounds
  sfx_trebrk, // tree breaking sound
  sfx_clratk, // cleric sounds
  sfx_clrpn,
  sfx_clrdth,
  sfx_cblsht, // cleric projectile sounds
  sfx_cblexp,
  sfx_gloop,  // water terrain sound
  sfx_thundr, // global lightning sound effect
  sfx_muck,   // more ambience, terrain sounds
  sfx_bubble,
  sfx_shlurp,
  sfx_rocks,
  sfx_clrdef, // leader cleric defense
  sfx_dfdth,  // dwarf sounds
  sfx_dwrfpn,
  sfx_dfsit1,
  sfx_dfsit2,
  sfx_fdsit,
  sfx_fdpn,
  sfx_fddth,
  sfx_heal,
  sfx_harp,
  sfx_wofp,
  sfx_nspdth, // new spider death
  sfx_chains,
  sfx_glass5,
  sfx_gong,
  sfx_lava2,
  sfx_locked,
  sfx_burn,
  sfx_mumhed,
  sfx_mumsit,
  sfx_mumpai,
  sfx_mumdth,
  sfx_amb1,
  sfx_cone3,
  sfx_icedth,
  sfx_plfeet,
  sfx_plfall,
  sfx_fallht,
  sfx_clsit1,
  sfx_clsit2,
  sfx_clsit3,
  sfx_clsit4,
  sfx_keys2a,
  // End Eternity TC New SFX
  sfx_buzz, // joel: ambient sounds
  sfx_comp,
  sfx_drip,
  sfx_fire,
  sfx_lava,
  sfx_river,
  sfx_wfall,
  sfx_wind,
  sfx_wind1,
  sfx_cross, // joel: holy cross sound
  NUMSFX
} sfxenum_t;

extern sfxinfo_t chgun;

#endif

//----------------------------------------------------------------------------
//
// $Log: sounds.h,v $
// Revision 1.3  1998/05/03  22:44:30  killough
// beautification
//
// Revision 1.2  1998/01/26  19:27:53  phares
// First rev with no ^Ms
//
// Revision 1.1.1.1  1998/01/19  14:03:03  rand
// Lee's Jan 19 sources
//
//----------------------------------------------------------------------------
